import typing

from starlette.datastructures import URL, Headers, MutableHeaders, QueryParams
from starlette.requests import Request
from starlette.responses import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    Response,
    StreamingResponse,
)

__all__ = [
    "Method",
    "Scheme",
    "Host",
    "Port",
    "Path",
    "QueryString",
    "QueryParam",
    "Header",
    "Body",
    "PathParams",
    "PathParam",
    "RequestData",
    "URL",
    "QueryParams",
    "Headers",
    "MutableHeaders",
    "Request",
    "Response",
    "PlainTextResponse",
    "HTMLResponse",
    "JSONResponse",
    "FileResponse",
    "RedirectResponse",
    "StreamingResponse",
    "ReturnValue",
]


Method = typing.NewType("Method", str)
Scheme = typing.NewType("Scheme", str)
Host = typing.NewType("Host", str)
Port = typing.NewType("Port", int)
Path = typing.NewType("Path", str)
QueryString = typing.NewType("QueryString", str)
QueryParam = typing.NewType("QueryParam", str)
Header = typing.NewType("Header", str)
Body = typing.NewType("Body", bytes)
PathParams = typing.NewType("PathParams", dict)
PathParam = typing.NewType("PathParam", str)
RequestData = typing.TypeVar("RequestData")
ReturnValue = typing.TypeVar("ReturnValue")



from flama import exceptions, websockets
from flama.codecs.base import WebsocketsCodec

__all__ = ["BytesCodec"]


class BytesCodec(WebsocketsCodec):
    encoding = "bytes"

    async def decode(self, message: websockets.Message, **options):
        if "bytes" not in message:
            raise exceptions.DecodeError("Expected bytes websocket messages")

        return message["bytes"]



import typing

import marshmallow
from starlette.responses import JSONResponse

__all__ = ["APIResponse", "APIErrorResponse", "APIError"]


class APIError(marshmallow.Schema):
    status_code = marshmallow.fields.Integer(title="status_code", description="HTTP status code", required=True)
    detail = marshmallow.fields.Raw(title="detail", description="Error detail", required=True)
    error = marshmallow.fields.String(title="type", description="Exception or error type")


class APIResponse(JSONResponse):
    media_type = "application/json"

    def __init__(self, schema: typing.Optional[marshmallow.Schema] = None, *args, **kwargs):
        self.schema = schema
        super().__init__(*args, **kwargs)

    def render(self, content: typing.Any):
        # Use output schema to validate and format data
        if self.schema is not None:
            content = self.schema.dump(content)

        return super().render(content)


class APIErrorResponse(APIResponse):
    def __init__(
        self, detail: typing.Any, status_code: int = 400, exception: typing.Optional[Exception] = None, *args, **kwargs
    ):
        content = {
            "status_code": status_code,
            "detail": detail,
            "error": str(exception.__class__.__name__) if exception is not None else None,
        }

        super().__init__(schema=APIError(), content=content, status_code=status_code, *args, **kwargs)

        self.detail = detail
        self.exception = exception



import asyncio
import functools
import inspect
import typing

from flama import http, websockets
from flama.components.asgi import ASGI_COMPONENTS, ASGIReceive, ASGIScope, ASGISend
from flama.components.validation import VALIDATION_COMPONENTS
from flama.exceptions import ComponentNotFound
from flama.routing import Route

__all__ = ["Injector"]


class Injector:
    def __init__(self, components):
        from flama.applications import Flama

        self.components = list(ASGI_COMPONENTS + VALIDATION_COMPONENTS) + components
        self.initial = {
            "scope": ASGIScope,
            "receive": ASGIReceive,
            "send": ASGISend,
            "exc": Exception,
            "app": Flama,
            "path_params": http.PathParams,
            "route": Route,
            "request": http.Request,
            "response": http.Response,
            "websocket": websockets.WebSocket,
            "websocket_message": websockets.Message,
            "websocket_encoding": websockets.Encoding,
            "websocket_code": websockets.Code,
        }
        self.reverse_initial = {val: key for key, val in self.initial.items()}
        self.resolver_cache = {}

    def resolve_parameter(
        self, parameter, kwargs: typing.Dict, consts: typing.Dict, seen_state: typing.Set, parent_parameter=None
    ) -> typing.List[typing.Tuple]:
        """
        Resolve a parameter by inferring the component that suits it or by adding a value to kwargs or consts.

        The list of steps returned consists of a resolver function, a boolean that indicates if the function is async,
        function kwargs and consts and the output name.

        :param parameter: parameter to be resolved.
        :param kwargs: kwargs that defines current context.
        :param consts: consts that defines current context.
        :param seen_state: cached state.
        :param parent_parameter: parent parameter.
        :return: list of steps to resolve the component.
        """
        if parameter.annotation is http.ReturnValue:
            kwargs[parameter.name] = "return_value"
            return []

        # Check if the parameter class exists in 'initial'.
        if parameter.annotation in self.reverse_initial:
            initial_kwarg = self.reverse_initial[parameter.annotation]
            kwargs[parameter.name] = initial_kwarg
            return []

        # The 'Parameter' annotation can be used to get the parameter
        # itself. Used for example in 'Header' components that need the
        # parameter name in order to lookup a particular value.
        if parameter.annotation is inspect.Parameter:
            consts[parameter.name] = parent_parameter
            return []

        for component in self.components:
            if component.can_handle_parameter(parameter):
                identity = component.identity(parameter)
                kwargs[parameter.name] = identity
                if identity not in seen_state:
                    seen_state.add(identity)
                    return self.resolve_component(
                        resolver=component.resolve,
                        output_name=identity,
                        seen_state=seen_state,
                        parent_parameter=parameter,
                    )

                return []
        else:
            raise ComponentNotFound(parameter.name)

    def resolve_component(
        self, resolver, output_name: str, seen_state: typing.Set, parent_parameter=None
    ) -> typing.List[typing.Tuple]:
        """
        Resolve a component injecting all dependencies needed in its resolver function.

        The list of steps returned consists of a resolver function, a boolean that indicates if the function is async,
        function kwargs and consts and the output name.

        :param resolver: component resolver function.
        :param output_name: name used for that component in application status.
        :param seen_state: cached status.
        :param parent_parameter: parent parameter.
        :return: list of steps to resolve the component.
        """
        signature = inspect.signature(resolver)

        steps = []
        kwargs = {}
        consts = {}

        if output_name is None:
            if signature.return_annotation in self.reverse_initial:
                output_name = self.reverse_initial[signature.return_annotation]
            else:
                output_name = "return_value"

        for parameter in signature.parameters.values():
            try:
                steps += self.resolve_parameter(
                    parameter, kwargs, consts, seen_state=seen_state, parent_parameter=parent_parameter
                )
            except ComponentNotFound:
                raise ComponentNotFound(parameter=parameter.name, resolver=resolver.__class__.__name__)

        is_async = asyncio.iscoroutinefunction(resolver)

        step = (resolver, is_async, kwargs, consts, output_name)
        steps.append(step)

        return steps

    def resolve(self, func) -> typing.Tuple[typing.Dict, typing.Dict, typing.List]:
        """
        Inspects a function and creates a resolution list of all components needed to run it. returning

        :param func: function to resolve.
        :return: the keyword arguments, consts for that function and the steps to resolve all components.
        """
        seen_state = set(self.initial)

        steps = []
        kwargs = {}
        consts = {}

        signature = inspect.signature(func)

        for parameter in signature.parameters.values():
            try:
                steps += self.resolve_parameter(parameter, kwargs, consts, seen_state=seen_state)
            except ComponentNotFound:
                raise ComponentNotFound(parameter=parameter.name, resolver=func.__name__)

        return kwargs, consts, steps

    async def inject(self, func, state: typing.Dict) -> typing.Callable:
        """
        Given a function, injects all components defined in its signature and returns the partialized function.

        :param func: function to be partialized.
        :param state: mapping of current application state to infer components state.
        :return: partialized function.
        """
        try:
            func_kwargs, func_consts, steps = self.resolver_cache[func]
        except KeyError:
            func_kwargs, func_consts, steps = self.resolve(func)
            self.resolver_cache[func] = (func_kwargs, func_consts, steps)

        for resolver, is_async, kwargs, consts, output_name in steps:
            kw = {key: state[val] for key, val in kwargs.items()}
            kw.update(consts)
            if is_async:
                state[output_name] = await resolver(**kw)
            else:
                state[output_name] = resolver(**kw)

        kw = {key: state[val] for key, val in func_kwargs.items()}
        kw.update(func_consts)

        return functools.partial(func, **kw)



import asyncio
import inspect
from functools import wraps

import marshmallow

from flama import exceptions

__all__ = ["get_output_schema", "output_validation"]


def get_output_schema(func):
    """
    Get output schema annotated as function's return. If there is no schema, return None.

    :param func: Annotated function.
    :returns: Output schema.
    """
    return_annotation = inspect.signature(func).return_annotation
    if inspect.isclass(return_annotation) and issubclass(return_annotation, marshmallow.Schema):
        return return_annotation()
    elif isinstance(return_annotation, marshmallow.Schema):
        return return_annotation

    return None


def output_validation(error_cls=exceptions.OutputValidationError, error_status_code=500):
    """
    Validate view output using schema annotated as function's return.

    :param error_cls: Error class to be raised when validation fails. Errors dict will be passed through 'detail' param.
    :param error_status_code: HTTP status code assigned to response when it fails to validate output, default 500.
    :raises exceptions.ValidationError: if output validation fails.
    """

    def outer_decorator(func):
        schema = get_output_schema(func)
        assert schema is not None, "Return annotation must be a valid marshmallow schema"

        @wraps(func)
        async def inner_decorator(*args, **kwargs):
            response = await func(*args, **kwargs) if asyncio.iscoroutinefunction(func) else func(*args, **kwargs)

            try:
                # Use output schema to first deserialize the date and later validate it
                errors = schema.validate(schema.dump(response))
                if errors:
                    raise error_cls(detail=errors, status_code=error_status_code)
            except marshmallow.ValidationError as e:
                raise error_cls(detail=e.messages, status_code=error_status_code)

            return response

        return inner_decorator

    return outer_decorator



from flama import http
from flama.codecs.base import HTTPCodec

__all__ = ["URLEncodedCodec"]


class URLEncodedCodec(HTTPCodec):
    media_type = "application/x-www-form-urlencoded"

    async def decode(self, request: http.Request, **options):
        return await request.form() or None



import datetime
import enum
import typing
import uuid

import marshmallow

try:
    from sqlalchemy import Table
except Exception:  # pragma: no cover
    Table = typing.Any

try:
    from databases import Database
except Exception:  # pragma: no cover
    Database = typing.Any

__all__ = [
    "FieldLocation",
    "Field",
    "EndpointInfo",
    "OptInt",
    "OptStr",
    "OptBool",
    "OptFloat",
    "OptUUID",
    "OptDate",
    "OptDateTime",
    "Model",
    "PrimaryKey",
    "ResourceMeta",
    "ResourceMethodMeta",
    "HTTPMethod",
]


class FieldLocation(enum.Enum):
    query = enum.auto()
    path = enum.auto()
    body = enum.auto()


class Field(typing.NamedTuple):
    name: str
    location: FieldLocation
    schema: typing.Union[marshmallow.fields.Field, marshmallow.Schema]
    required: bool = False


class EndpointInfo(typing.NamedTuple):
    path: str
    method: str
    func: typing.Callable
    query_fields: typing.Dict[str, Field]
    path_fields: typing.Dict[str, Field]
    body_field: Field
    output_field: typing.Any


OptInt = typing.Optional[int]
OptStr = typing.Optional[str]
OptBool = typing.Optional[bool]
OptFloat = typing.Optional[float]
OptUUID = typing.Optional[uuid.UUID]
OptDate = typing.Optional[datetime.date]
OptDateTime = typing.Optional[datetime.datetime]


class PrimaryKey(typing.NamedTuple):
    name: str
    type: typing.Any


class Model(typing.NamedTuple):
    table: Table
    primary_key: PrimaryKey


class ResourceMeta(typing.NamedTuple):
    model: Model
    input_schema: marshmallow.Schema
    output_schema: marshmallow.Schema
    database: Database
    name: str
    verbose_name: str
    columns: typing.Sequence[str]
    order: str


class ResourceMethodMeta(typing.NamedTuple):
    path: str
    methods: typing.List[str] = ["GET"]
    name: str = None
    kwargs: typing.Dict[str, typing.Any] = {}


HTTPMethod = enum.Enum("HTTPMethod", ["GET", "HEAD", "POST", "PUT", "DELETE", "CONNECT", "OPTIONS", "TRACE", "PATCH"])



import datetime
import logging
import re
import typing
import uuid

import marshmallow

from flama.exceptions import HTTPException
from flama.pagination import Paginator
from flama.responses import APIResponse
from flama.types import Model, PrimaryKey, ResourceMeta, ResourceMethodMeta

try:
    import sqlalchemy
    from sqlalchemy.dialects import postgresql
except Exception:  # pragma: no cover
    raise AssertionError("`sqlalchemy` must be installed to use resources") from None

try:
    import databases
except Exception:  # pragma: no cover
    raise AssertionError("`databases` must be installed to use resources") from None


logger = logging.getLogger(__name__)


__all__ = ["resource_method", "BaseResource", "CRUDResource", "CRUDListResource", "CRUDListDropResource"]


PK_MAPPING = {
    sqlalchemy.Integer: int,
    sqlalchemy.String: str,
    sqlalchemy.Date: datetime.date,
    sqlalchemy.DateTime: datetime.datetime,
    postgresql.UUID: uuid.UUID,
}


class DropCollection(marshmallow.Schema):
    deleted = marshmallow.fields.Integer(title="deleted", description="Number of deleted elements", required=True)


def resource_method(path: str, methods: typing.List[str] = None, name: str = None, **kwargs) -> typing.Callable:
    def wrapper(func: typing.Callable) -> typing.Callable:
        func._meta = ResourceMethodMeta(
            path=path, methods=methods if methods is not None else ["GET"], name=name, kwargs=kwargs
        )

        return func

    return wrapper


class ResourceRoutes:
    """Routes descriptor"""

    def __init__(self, methods: typing.Dict[str, typing.Callable]):
        self.methods = methods

    def __get__(self, instance, owner) -> typing.Dict[str, typing.Callable]:
        return self.methods


class BaseResource(type):
    METHODS = ()  # type: typing.Sequence[str]

    def __new__(mcs, name, bases, namespace):
        # Get database and replace it with a read-only descriptor
        database = mcs._get_attribute("database", name, namespace, bases)
        namespace["database"] = property(lambda self: self._meta.database)

        # Get model and replace it with a read-only descriptor
        model = mcs._get_model(name, namespace, bases)
        namespace["model"] = property(lambda self: self._meta.model.table)

        # Define resource names
        resource_name, verbose_name = mcs._get_resource_name(name, namespace)

        # Default columns and order for admin interface
        columns = namespace.pop("columns", [model.primary_key.name])
        order = namespace.pop("order", model.primary_key.name)

        # Get input and output schemas
        input_schema, output_schema = mcs._get_schemas(name, namespace, bases)

        namespace["_meta"] = ResourceMeta(
            database=database,
            model=model,
            name=resource_name,
            verbose_name=verbose_name,
            input_schema=input_schema,
            output_schema=output_schema,
            columns=columns,
            order=order,
        )

        # Create CRUD methods and routes
        mcs._add_methods(resource_name, verbose_name, namespace, database, input_schema, output_schema, model)
        mcs._add_routes(namespace)

        return super().__new__(mcs, name, bases, namespace)

    @classmethod
    def _get_attribute(
        mcs, attribute: str, name: str, namespace: typing.Dict[str, typing.Any], bases: typing.Sequence[typing.Any]
    ) -> typing.Any:
        try:
            return namespace.pop(attribute)
        except KeyError:
            for base in bases:
                if hasattr(base, "_meta") and hasattr(base._meta, attribute):
                    return getattr(base._meta, attribute)
                elif hasattr(base, attribute):
                    return getattr(base, attribute)

        raise AttributeError(f"{name} needs to define attribute '{attribute}'")

    @classmethod
    def _get_resource_name(mcs, name: str, namespace: typing.Dict[str, typing.Any]) -> typing.Tuple[str, str]:
        resource_name = namespace.pop("name", name.lower())

        # Check resource name validity
        if re.match("[a-zA-Z][-_a-zA-Z]", resource_name) is None:
            raise AttributeError(f"Invalid resource name '{resource_name}'")

        return resource_name, namespace.pop("verbose_name", resource_name)

    @classmethod
    def _get_model(
        mcs, name: str, namespace: typing.Dict[str, typing.Any], bases: typing.Sequence[typing.Any]
    ) -> Model:
        model = mcs._get_attribute("model", name, namespace, bases)

        # Already defined model probably because resource inheritance, so no need to create it
        if isinstance(model, Model):
            return model

        # Resource define model as a sqlalchemy Table, so extract necessary info from it
        elif isinstance(model, sqlalchemy.Table):
            # Get model primary key
            model_pk = list(sqlalchemy.inspect(model).primary_key.columns.values())

            # Check primary key exists and is a single column
            if len(model_pk) != 1:
                raise AttributeError(f"{name} model must define a single-column primary key")

            model_pk = model_pk[0]
            model_pk_name = model_pk.name

            # Check primary key is a valid type
            try:
                model_pk_type = PK_MAPPING[model_pk.type.__class__]
            except KeyError:
                raise AttributeError(
                    f"{name} model primary key must be any of {', '.join((i.__name__ for i in PK_MAPPING.keys()))}"
                )

            return Model(table=model, primary_key=PrimaryKey(model_pk_name, model_pk_type))

        raise AttributeError(f"{name} model must be a valid SQLAlchemy Table instance or a Model instance")

    @classmethod
    def _get_schemas(
        mcs, name: str, namespace: typing.Dict[str, typing.Any], bases: typing.Sequence[typing.Any]
    ) -> typing.Tuple[marshmallow.Schema, marshmallow.Schema]:
        try:
            schema = mcs._get_attribute("schema", name, namespace, bases)
            input_schema = schema
            output_schema = schema
        except AttributeError:
            try:
                input_schema = mcs._get_attribute("input_schema", name, namespace, bases)
                output_schema = mcs._get_attribute("output_schema", name, namespace, bases)
            except AttributeError:
                raise AttributeError(
                    f"{name} needs to define attribute 'schema' or the pair 'input_schema' and 'output_schema'"
                )

        return input_schema, output_schema

    @classmethod
    def _add_routes(mcs, namespace: typing.Dict[str, typing.Any]):
        methods = {name: m for name, m in namespace.items() if getattr(m, "_meta", False) and not name.startswith("_")}
        routes = ResourceRoutes(methods)

        namespace["routes"] = routes

    @classmethod
    def _add_methods(
        mcs,
        name: str,
        verbose_name: str,
        namespace: typing.Dict[str, typing.Any],
        database: "databases.Database",
        input_schema: marshmallow.Schema,
        output_schema: marshmallow.Schema,
        model: Model,
    ):
        # Get available methods
        methods = [getattr(mcs, f"_add_{method}") for method in mcs.METHODS if hasattr(mcs, f"_add_{method}")]

        # Generate CRUD methods
        crud_namespace = {
            func_name: func
            for method in methods
            for func_name, func in method(
                name=name,
                verbose_name=verbose_name,
                database=database,
                input_schema=input_schema,
                output_schema=output_schema,
                model=model,
            ).items()
        }

        # Preserve already defined methods
        crud_namespace.update(
            {method: crud_namespace[f"_{method}"] for method in mcs.METHODS if method not in namespace}
        )

        namespace.update(crud_namespace)


class CreateMixin:
    @classmethod
    def _add_create(
        mcs,
        name: str,
        verbose_name: str,
        database: databases.Database,
        input_schema: marshmallow.Schema,
        output_schema: marshmallow.Schema,
        **kwargs,
    ) -> typing.Dict[str, typing.Any]:
        @resource_method("/", methods=["POST"], name=f"{name}-create")
        @database.transaction()
        async def create(self, element: input_schema) -> output_schema:
            query = self.model.insert().values(**element)
            await self.database.execute(query)
            return APIResponse(schema=output_schema(), content=element, status_code=201)

        create.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Create a new document.
            description:
                Create a new document in this resource.
            responses:
                201:
                    description:
                        Document created successfully.
        """

        return {"_create": create}


class RetrieveMixin:
    @classmethod
    def _add_retrieve(
        mcs, name: str, verbose_name: str, output_schema: marshmallow.Schema, model: Model, **kwargs
    ) -> typing.Dict[str, typing.Any]:
        @resource_method("/{element_id}/", methods=["GET"], name=f"{name}-retrieve")
        async def retrieve(self, element_id: model.primary_key.type) -> output_schema:
            query = self.model.select().where(self.model.c[model.primary_key.name] == element_id)
            element = await self.database.fetch_one(query)

            if element is None:
                raise HTTPException(status_code=404)

            return dict(element)

        retrieve.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Retrieve a document.
            description:
                Retrieve a document from this resource.
            responses:
                200:
                    description:
                        Document found.
                404:
                    description:
                        Document not found.
        """

        return {"_retrieve": retrieve}


class UpdateMixin:
    @classmethod
    def _add_update(
        mcs,
        name: str,
        verbose_name: str,
        database: databases.Database,
        input_schema: marshmallow.Schema,
        output_schema: marshmallow.Schema,
        model: Model,
        **kwargs,
    ) -> typing.Dict[str, typing.Any]:
        @resource_method("/{element_id}/", methods=["PUT"], name=f"{name}-update")
        @database.transaction()
        async def update(self, element_id: model.primary_key.type, element: input_schema) -> output_schema:
            query = sqlalchemy.select([sqlalchemy.exists().where(self.model.c[model.primary_key.name] == element_id)])
            exists = next((i for i in (await self.database.fetch_one(query)).values()))
            if not exists:
                raise HTTPException(status_code=404)

            query = self.model.update().where(self.model.c[model.primary_key.name] == element_id).values(**element)
            await self.database.execute(query)

            return {model.primary_key.name: element_id, **element}

        update.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Update a document.
            description:
                Update a document in this resource.
            responses:
                200:
                    description:
                        Document updated successfully.
                404:
                    description:
                        Document not found.
        """

        return {"_update": update}


class DeleteMixin:
    @classmethod
    def _add_delete(
        mcs, name: str, verbose_name: str, database: databases.Database, model: Model, **kwargs
    ) -> typing.Dict[str, typing.Any]:
        @resource_method("/{element_id}/", methods=["DELETE"], name=f"{name}-delete")
        @database.transaction()
        async def delete(self, element_id: model.primary_key.type):
            query = sqlalchemy.select([sqlalchemy.exists().where(self.model.c[model.primary_key.name] == element_id)])
            exists = next((i for i in (await self.database.fetch_one(query)).values()))
            if not exists:
                raise HTTPException(status_code=404)

            query = self.model.delete().where(self.model.c[model.primary_key.name] == element_id)
            await self.database.execute(query)

            return APIResponse(status_code=204)

        delete.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Delete a document.
            description:
                Delete a document in this resource.
            responses:
                204:
                    description:
                        Document deleted successfully.
                404:
                    description:
                        Document not found.
        """

        return {"_delete": delete}


class ListMixin:
    @classmethod
    def _add_list(
        mcs, name: str, verbose_name: str, output_schema: marshmallow.Schema, **kwargs
    ) -> typing.Dict[str, typing.Any]:
        async def filter(self, *clauses, **filters) -> typing.List[typing.Dict]:
            query = self.model.select()

            where_clauses = tuple(clauses) + tuple(self.model.c[k] == v for k, v in filters.items())

            if where_clauses:
                query = query.where(sqlalchemy.and_(*where_clauses))

            return [dict(row) for row in await self.database.fetch_all(query)]

        @resource_method("/", methods=["GET"], name=f"{name}-list")
        @Paginator.page_number
        async def list(self, **kwargs) -> output_schema(many=True):
            return await self._filter()  # noqa

        list.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                List collection.
            description:
                List resource collection.
            responses:
                200:
                    description:
                        List collection items.
        """

        return {"_list": list, "_filter": filter}


class DropMixin:
    @classmethod
    def _add_drop(
        mcs, name: str, verbose_name: str, database: databases.Database, model: Model, **kwargs
    ) -> typing.Dict[str, typing.Any]:
        @resource_method("/", methods=["DELETE"], name=f"{name}-drop")
        @database.transaction()
        async def drop(self) -> DropCollection:
            query = sqlalchemy.select([sqlalchemy.func.count(self.model.c[model.primary_key.name])])
            result = next((i for i in (await self.database.fetch_one(query)).values()))

            query = self.model.delete()
            await self.database.execute(query)

            return APIResponse(schema=DropCollection(), content={"deleted": result}, status_code=204)

        drop.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Drop collection.
            description:
                Drop resource collection.
            responses:
                204:
                    description:
                        Collection dropped successfully.
        """

        return {"_drop": drop}


class CRUDResource(BaseResource, CreateMixin, RetrieveMixin, UpdateMixin, DeleteMixin):
    METHODS = ("create", "retrieve", "update", "delete")


class CRUDListResource(BaseResource, CreateMixin, RetrieveMixin, UpdateMixin, DeleteMixin, ListMixin):
    METHODS = ("create", "retrieve", "update", "delete", "list")


class CRUDListDropResource(BaseResource, CreateMixin, RetrieveMixin, UpdateMixin, DeleteMixin, ListMixin, DropMixin):
    METHODS = ("create", "retrieve", "update", "delete", "list", "drop")



import typing

import marshmallow

from flama.responses import APIResponse

__all__ = ["PageNumberSchema", "PageNumberResponse"]


class PageNumberMeta(marshmallow.Schema):
    page = marshmallow.fields.Integer(title="page", description="Current page number")
    page_size = marshmallow.fields.Integer(title="page_size", description="Page size")
    count = marshmallow.fields.Integer(title="count", description="Total number of items", allow_none=True)


class PageNumberSchema(marshmallow.Schema):
    meta = marshmallow.fields.Nested(PageNumberMeta)
    data = marshmallow.fields.List(marshmallow.fields.Dict())


class PageNumberResponse(APIResponse):
    """
    Response paginated based on a page number and a page size.

    First 10 elements:
        /resource?page=1
    Third 10 elements:
        /resource?page=3
    First 20 elements:
        /resource?page=1&page_size=20
    """

    default_page_size = 10

    def __init__(
        self,
        schema: marshmallow.Schema,
        page: typing.Optional[typing.Union[int, str]] = None,
        page_size: typing.Optional[typing.Union[int, str]] = None,
        count: typing.Optional[bool] = True,
        **kwargs
    ):
        self.page_number = int(page) if page is not None else 1
        self.page_size = int(page_size) if page_size is not None else self.default_page_size
        self.count = count
        super().__init__(schema=schema, **kwargs)

    def render(self, content: typing.Sequence):
        init = (self.page_number - 1) * self.page_size
        end = self.page_number * self.page_size

        return super().render(
            {
                "meta": {
                    "page": self.page_number,
                    "page_size": self.page_size,
                    "count": len(content) if self.count else None,
                },
                "data": content[init:end],
            }
        )



from flama import http
from flama.codecs.base import HTTPCodec

__all__ = ["MultiPartCodec"]


class MultiPartCodec(HTTPCodec):
    media_type = "multipart/form-data"

    async def decode(self, request: http.Request, **options):
        return await request.form()



import typing

__all__ = ["dict_safe_add"]


def dict_safe_add(d: typing.Dict, v: typing.Any, *keys):
    _d = d

    for k in keys[:-1]:
        if not _d.get(k):
            _d[k] = {}

        _d = _d[k]

    _d[keys[-1]] = v



import typing
from inspect import Parameter
from urllib.parse import parse_qsl

from flama import http
from flama.components import Component

ASGIScope = typing.NewType("ASGIScope", dict)
ASGIReceive = typing.NewType("ASGIReceive", typing.Callable)
ASGISend = typing.NewType("ASGISend", typing.Callable)


class MethodComponent(Component):
    def resolve(self, scope: ASGIScope) -> http.Method:
        return http.Method(scope["method"])


class URLComponent(Component):
    def resolve(self, scope: ASGIScope) -> http.URL:
        return http.URL(scope=scope)


class SchemeComponent(Component):
    def resolve(self, scope: ASGIScope) -> http.Scheme:
        return http.Scheme(scope["scheme"])


class HostComponent(Component):
    def resolve(self, scope: ASGIScope) -> http.Host:
        return http.Host(scope["server"][0])


class PortComponent(Component):
    def resolve(self, scope: ASGIScope) -> http.Port:
        return http.Port(scope["server"][1])


class PathComponent(Component):
    def resolve(self, scope: ASGIScope) -> http.Path:
        return http.Path(scope.get("root_path", "") + scope["path"])


class QueryStringComponent(Component):
    def resolve(self, scope: ASGIScope) -> http.QueryString:
        return http.QueryString(scope["query_string"].decode())


class QueryParamsComponent(Component):
    def resolve(self, scope: ASGIScope) -> http.QueryParams:
        query_string = scope["query_string"].decode()
        return http.QueryParams(parse_qsl(query_string))


class QueryParamComponent(Component):
    def resolve(self, parameter: Parameter, query_params: http.QueryParams) -> http.QueryParam:
        name = parameter.name
        if name not in query_params:
            return None
        return http.QueryParam(query_params[name])


class HeadersComponent(Component):
    def resolve(self, scope: ASGIScope) -> http.Headers:
        return http.Headers(scope=scope)


class HeaderComponent(Component):
    def resolve(self, parameter: Parameter, headers: http.Headers) -> http.Header:
        name = parameter.name.replace("_", "-")
        if name not in headers:
            return None
        return http.Header(headers[name])


class BodyComponent(Component):
    async def resolve(self, receive: ASGIReceive) -> http.Body:
        body = b""
        while True:
            message = await receive()
            if not message["type"] == "http.request":
                raise Exception(f"Unexpected ASGI message type '{message['type']}'.")
            body += message.get("body", b"")
            if not message.get("more_body", False):
                break

        return http.Body(body)


ASGI_COMPONENTS = (
    MethodComponent(),
    URLComponent(),
    SchemeComponent(),
    HostComponent(),
    PortComponent(),
    PathComponent(),
    QueryStringComponent(),
    QueryParamsComponent(),
    QueryParamComponent(),
    HeadersComponent(),
    HeaderComponent(),
    BodyComponent(),
)



import typing

from starlette.websockets import Message, WebSocket

__all__ = ["WebSocket", "Message", "Code", "Encoding", "Data"]


Code = typing.NewType("Code", int)
Encoding = typing.NewType("Encoding", str)
Data = typing.TypeVar("Data")



from flama.codecs.websockets.bytes import BytesCodec
from flama.codecs.websockets.json import JSONCodec
from flama.codecs.websockets.text import TextCodec

__all__ = ["BytesCodec", "TextCodec", "JSONCodec"]



import typing

from starlette.applications import Starlette
from starlette.exceptions import ExceptionMiddleware
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.types import ASGIApp

from flama import exceptions
from flama.components import Component
from flama.exceptions import HTTPException
from flama.http import Request, Response
from flama.injection import Injector
from flama.responses import APIErrorResponse
from flama.routing import Router
from flama.schemas import SchemaMixin

if typing.TYPE_CHECKING:
    from flama.resources import BaseResource

__all__ = ["Flama"]


class Flama(Starlette, SchemaMixin):
    def __init__(
        self,
        components: typing.Optional[typing.List[Component]] = None,
        debug: bool = False,
        title: typing.Optional[str] = "",
        version: typing.Optional[str] = "",
        description: typing.Optional[str] = "",
        schema: typing.Optional[str] = "/schema/",
        docs: typing.Optional[str] = "/docs/",
        redoc: typing.Optional[str] = None,
        *args,
        **kwargs
    ) -> None:
        super().__init__(debug=debug, *args, **kwargs)

        if components is None:
            components = []

        # Initialize injector
        self.components = components

        self.router = Router(components=components)
        self.app = self.router
        self.exception_middleware = ExceptionMiddleware(self.router, debug=debug)
        self.error_middleware = ServerErrorMiddleware(self.exception_middleware, debug=debug)

        # Add exception handler for API exceptions
        self.add_exception_handler(exceptions.HTTPException, self.api_http_exception_handler)

        # Add schema and docs routes
        self.add_schema_docs_routes(
            title=title, version=version, description=description, schema=schema, docs=docs, redoc=redoc
        )

    @property
    def injector(self):
        return Injector(components=self.components)

    def mount(self, path: str, app: ASGIApp, name: str = None) -> None:
        self.components += getattr(app, "components", [])
        self.router.mount(path, app=app, name=name)

    def add_resource(self, path: str, resource: "BaseResource"):
        self.router.add_resource(path, resource=resource)

    def resource(self, path: str) -> typing.Callable:
        def decorator(resource: "BaseResource") -> "BaseResource":
            self.router.add_resource(path, resource=resource)
            return resource

        return decorator

    def api_http_exception_handler(self, request: Request, exc: HTTPException) -> Response:
        return APIErrorResponse(detail=exc.detail, status_code=exc.status_code, exception=exc)
